from zope.interface import implements
from ironman.interfaces import IHardwareManager, IHardwareMap, IHardwareNode
import yaml

class HardwareManager(dict):
    implements(IHardwareManager)
    raw_maps = {}

    def check_data(self, address, data):
        node = self.get_node(address)
        return node.isValueValid(data)

    def check_address(self, address):
        return not isinstance(self.get_node(address), NullHardwareNode)

    def find_address(self, address):
        return self.get(address, NullHardwareNode())

    def get_route(self, address):
        return self.get_node(address).hw_map.route

    def get_node(self, address):
        return self.get(address, NullHardwareNode())

    def get_checksum(self, map_name):
        pass

    def add(self, new_hw_map):
        """
            Add the HW map only if it doesn't exist for a given key, and no address collisions
        """
        new_route = new_hw_map.route
        if new_route in self.raw_maps:
            raise KeyError("HW Map already exists: {0:s}".format(new_hw_map.route))

        common_addresses = set(self).intersection(new_hw_map)
        if common_addresses:
            raise ValueError("An address in {0:s} already exists in the manager".format(new_route))
        # all ok, add it all
        self.raw_maps[new_route] = new_hw_map
        self.update(new_hw_map)

    def subtract(self, route):
        """
            Remove the route entirely.
        """
        for address in self.raw_maps.pop(route, NullHardwareMap()).iterkeys():
            self.pop(address, NullHardwareNode())

class NullHardwareMap(dict):
    implements(IHardwareMap)
    route = None
    def parse(self, yml): pass
    def isOk(self): return False

class HardwareMap(dict):
    implements(IHardwareMap)

    def __init__(self, yml, route):
        self.route = route
        self.parse(yml)

    def parse(self, yml):
        doc = yaml.load(yml)
        for node in doc.get('nodes', []):
            baseAddress = node.get('address')
            # this will check if there are any children later
            child = None
            for child in node.get('nodes', []):
                childAddress = child.get('address')
                absAddress = baseAddress+childAddress
                if absAddress in self: raise KeyError('{0:s}/{1:s}'.format(node['id'],child['id']))
                self[absAddress] = HardwareNode(child, self)
            # no children
            if child is None:
                self[baseAddress] = HardwareNode(node, self)

    def isOk(self):
        for k,v in self.iteritems():
            if not v.isOk(): return False
        return True

class NullHardwareNode(dict):
    implements(IHardwareNode)

    hw_map = NullHardwareMap()
    readable = False
    writeable = False
    isValueValid = False
    isOk = False
    permissions = set()
    allowed = set()
    disallowed = set()

class HardwareNode(dict):
    implements(IHardwareNode)

    def __init__(self, node, hw_map):
        #self['description'] = getattr(node, 'description', '')
        self['permissions'] = int(node.get('permissions', 0))
        self['allowed'] = node.get('allowed', [])
        self['disallowed'] = node.get('disallowed', ['-1'])
        self.hw_map = hw_map

    @property
    def readable(self):
        return bool(self['permissions']&1)

    @property
    def writeable(self):
        return bool(self['permissions']&2)

    def isValueValid(self, val):
        return (self['allowed'] and val in self['allowed']) or \
               (self['disallowed'] and val not in self['disallowed'])

    @property
    def isOk(self):
        return (bool(self['allowed'])^bool(self['disallowed'])) and (self['permissions']&1)

    @property
    def permissions(self):
        return self['permissions']

    @property
    def allowed(self):
        return self['allowed']

    @property
    def disallowed(self):
        return self['disallowed']



from ironman.hardware import HardwareManager, HardwareMap, HardwareNode
manager = HardwareManager()
manager.add(HardwareMap(file('xadc.yml').read(), 'xadc'))

from ironman.communicator import Jarvis, ComplexIO
j = Jarvis()
j.set_hardware_manager(manager)

@j.register('xadc')
class XADCController(ComplexIO):
    __base__ = "/sys/devices/soc0/amba/f8007100.adc/iio:device0/"
    __f__ = {
                0:   __base__+"in_temp0_offset",
                1:   __base__+"in_temp0_raw",
                2:   __base__+"in_temp0_scale",
                17:  __base__+"in_voltage0_vccint_raw",
                18:  __base__+"in_voltage0_vccint_scale",
                33:  __base__+"in_voltage1_vccaux_raw",
                34:  __base__+"in_voltage1_vccaux_scale",
                49:  __base__+"in_voltage2_vccbram_raw",
                50:  __base__+"in_voltage2_vccbram_scale",
                65:  __base__+"in_voltage3_vccpint_raw",
                66:  __base__+"in_voltage3_vccpint_scale",
                81:  __base__+"in_voltage4_vccpaux_raw",
                82:  __base__+"in_voltage4_vccpaux_scale",
                97:  __base__+"in_voltage5_vccoddr_raw",
                98:  __base__+"in_voltage5_vccoddr_scale",
                113: __base__+"in_voltage6_vrefp_raw",
                114: __base__+"in_voltage6_vrefp_scale",
                129: __base__+"in_voltage7_vrefn_raw",
                130: __base__+"in_voltage7_vrefn_scale"
            }

fakereaderyaml = '''
nodes:
    -
        id: random
        address: 0x00000003
'''
manager.add(HardwareMap(fakereaderyaml, 'fake'))

import random
@j.register('fake')
class FakeReader:
    def read(self, offset, size):
        number = random.random()
        return str(number)[:size*4]

    def write(self, offset, data): pass

from ironman.constructs.ipbus import PacketHeaderStruct, ControlHeaderStruct, IPBusConstruct, IPBusWords
def buildResponsePacket(packet):
    packet.response.data[0].info_code = 'SUCCESS'
    #packet.response.data[0].data = [packet.response.data[0].data]
    return IPBusConstruct.build(packet.response)
    # data += PacketHeaderStruct.build(packet.response.header)
    # for transaction, response in zip(packet.response.data, packet.response):
    #     data += ControlHeaderStruct.build(transaction)
    #     data += response.encode("hex").decode("hex")
    return data


from ironman.history import History
h = History()

from ironman.server import ServerFactory
from ironman.packet import IPBusPacket
from twisted.internet import reactor
from twisted.internet.defer import Deferred

def deferredGenerator():
    return Deferred().addCallback(IPBusPacket).addCallback(j).addCallback(buildResponsePacket)#.addCallback(h.record)

reactor.listenUDP(8888, ServerFactory('udp', deferredGenerator))

'''set up a web server from top level: http://imgur.com/jCQlfrm'''
# Site, an IProtocolFactory which glues a listening server port (IListeningPort) to the HTTPChannel implementation
from twisted.web.server import Site
# File, an IResource which glues the HTTP protocol implementation to the filesystem
from twisted.web.static import File
reactor.listenTCP(8000, Site(File("/")))

'''set up a mirror web server for IPBus requests'''
from twisted.web.resource import Resource
# deferred responses
from twisted.web.server import NOT_DONE_YET
# return json
import json
class HTTPIPBusRoot(Resource):
    # has children
    isLeaf = False

class HTTPIPBus(Resource):
    # no children
    isLeaf = True
    def render_GET(self, request):
        request.responseHeaders.addRawHeader(b"content-type", b"application/json")
        # request.postpath will contain what we need
        if len(request.postpath) != 3:
            return json.dumps({
                "success": False,
                "data": None,
                "error": "Incorrect number of URL segments",
                "traceback": None
            })

        # make sure we can convert to hex string
        for i, word in enumerate(request.postpath):
            if len(word) != 8:
                return json.dumps({
                    "success": False,
                    "data": None,
                    "error": "Word {:d} has incorrect length {:d}. It should be: 8.".format(i, len(word)),
                    "traceback": None})

        packet = ''.join(request.postpath)
        try:
            packet.decode("hex")
        except TypeError as e:
            import sys
            return json.dumps({
                "success": False,
                "data": None,
                "error": str(e),
                "traceback": str(sys.exc_info())
            })

        def write(result):
            request.write(json.dumps({
                "success": True,
                "data": IPBusWords.build(IPBusConstruct.parse(result).data[0]),
                "error": None,
                "traceback": None
            }))
            request.finish()

        def error(result):
            request.write(json.dumps({
                "success": False,
                "data": None,
                "error": "An unknown error has occurred with the application. Message: {0}".format(result.getErrorMessage()),
                "traceback": result.getBriefTraceback()
            }))
            request.finish()

        d = deferredGenerator()
        d.addCallbacks(write, error)
        print("Handling IPBus packet {0:s}".format(packet))
        d.callback(packet.decode('hex'))
        return NOT_DONE_YET

http_ipbus_root = HTTPIPBusRoot()
http_ipbus_root.putChild('read', HTTPIPBus())
reactor.listenTCP(7777, Site(http_ipbus_root))

reactor.run()



__version__ = '0.2.17'
__all__ = ['communicator',
           'hardware',
           'history',
           'interfaces',
           'packet',
           'server',
           'utilities']

def engage(proto='udp'):
    ''' Fire thrusters.
    '''
    from ironman.server import ServerFactory
    from twisted.internet import reactor
    from twisted.internet.defer import Deferred
    getattr(reactor, 'listen{0:s}'.format(proto.upper()))(8888, ServerFactory(proto, Deferred))
    reactor.run()



#!/usr/bin/env python
from __future__ import print_function
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import io, os, sys

import ironman

here = os.path.abspath(os.path.dirname(__file__))

def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)

long_description = read(os.path.join(here, 'README.rst'))#, 'CHANGES.txt')

class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['--strict', '--verbose', '--tb=long', 'tests']
        self.test_suite = True

    def run_tests(self):
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)

setup(name='ironman',
      version=ironman.__version__,
      url='https://github.com/kratsg/ironman',
      license='MIT',
      description='SoC Firmware for L1Calo',
      long_description=long_description,
      author='Giordon Stark',
      author_email='gstark@cern.ch',
      maintainer='Giordon Stark',
      maintainer_email='gstark@cern.ch',
      tests_require=['pytest'],
      install_requires=['Twisted>=15.4.0',
                        'zope.interface==4.1.3',
                        'construct==2.5.2',
                        'PyYAML==3.11'],
      cmdclass={'test': PyTest},
      packages=find_packages(),
      test_suite='ironman.test.test_ironman',
      classifiers=[
        'Programming Language :: Python',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)'
      ],
      extras_require={
        'testing': ['pytest'],
      },
      data_files=[('workbench', ['workbench/gFEXTest.py', 'workbench/xadc.yml'])],
)



from zope.interface import Interface, Attribute

class IHistory(Interface):
    """ Enhanced dictionary to store inbound and outbound packet pairs
    """

    def record(packet):
        """ record the packet
        """

class IIPBusPacket(Interface):
    """ IPBus Packet object
    """
    request = Attribute("The parsed request packet")
    response = Attribute("The parsed response packet")
    _raw = Attribute("The raw request packet")
    raw = Attribute("The raw datagram blob.")
    littleendian = Attribute("A flag dictating whether the datagram is received/sent in little-endian.")
    protocol_version = Attribute("The packet header protocol version. This does not check that the encapsulated transactions also match.")
    reserved = Attribute("Reserved. Should be 0x0.")
    packet_id = Attribute("The id of the ipbus packet.")
    byteorder = Attribute("The byte-order in the header. Should assert == 0xf.")
    packet_type = Attribute("""The type of packet.

            ======= ===============
            Value   Type
            ======= ===============
            0x0     Control
            0x1     Status
            0x2     Re-send request
            0x3-f   Reserved
            ======= ===============""")
    response = Attribute("The data response to be passed along to another function that builds the response packet. This should be a list [] to append responses to.")


    def __init__(blob):
        """ Packet is initialized with a data blob
            to decode. Determine if it is big or little endian.
        """

    def __eq__(other):
        """ Define a way to identify two packets as being equivalent. Best way is to compare the underlying structs
        """

    def __ne__(other):
        """ This should just be :code:`return not self.__eq__(other)`.
        """

class IHardwareManager(Interface):
    """ Our Hardware Maps manager
    """

    raw_maps = Attribute("A dictionary of the maps added so we can keep track which makes it easier to add and remove.")

    def get_node(address):
        """ Given an address, return the node associated with it
        """

    def get_route(address):
        """ Given an address, return the route for it
        """

    def get_checksum(route):
        """ Look up the checksum for a given map name (route)
        """

    def check_address(address):
        """ Given an address, checks if it is valid
        """

    def check_data(address, data):
        """ Given an address, checks if the data is
            a valid value to write
        """

    def add(hw_map):
        """ Add the Map object to the Hardware
        """

    def subtract(route):
        """ Remove the route from the hardware manager
        """

class IHardwareMap(Interface):
    """ Manages information about a single map, should be an overloaded dictionary
    """
    route = Attribute("The route associated for this hardware map.")

    def __init__(xml, route):
        """ Initialize a hardware map object by giving it the data to parse and associate it with a route
        """

    def parse(xml):
        """ Parse the xml hardware map data to set things up
        """

    def isOk():
        """ Whether or not the given hardware map is ok. Should just be a loop over :func:`IHardwareNode.isOk`.
        """

class IHardwareNode(Interface):
    """ Manages information about a single address. Simply a well-defined dictionary.
    """
    #description = Attribute("A description of the node.")
    permissions = Attribute("Mark the node's read/write capabilities.")
    allowed = Attribute("A list of allowed values for the node.")
    disallowed = Attribute("A list of disallowed values for the node.")
    readable = Attribute("Is the given node readable?")
    writeable = Attribute("Is the given node writeable?")
    isOk = Attribute("Is the given node ok? EG: can't set allowed and disallowed objects at the same time and cannot block a node from being readable.")
    hw_map = Attribute("The hardware map this is associated with.")

    def __init__(node, hw_map):
        """ Initialize the node by giving it the parsed xml data as well as the hw_map
        """

    def isValueValid(val):
        """ Whether the given value is a valid value for the node
        """

class ICommunicationSlave(Interface):
    """ Manages the communication with the programmable logic for us
    """
    def set_hardware_manager(hwmanager):
        """ Set the hardware manager that the slave communications with
        """

    def parse_address(address):
        """ Parses address and returns what function to call
        """

    def __call__(packet):
        """ A non-blocking I/O call passing along the packet

            Returns the responses
        """

    def __transaction__(transaction):
        """ Handle a single transaction and return the response
        """

class ICommunicationDriver(Interface):
    """ The standard driver that is expected for all methods of communication on the board
    """
    def read(offset, size):
        """ Read from the given address (offset) for N bytes (size)
        """

    def write(offset, value):
        """ Write to the given address (offset) for N bytes (len(value))
        """
